"""
Useful functions for data reduction.
"""
import copy
import sys
import warnings

import numpy as np


def __init_pipeline_outputs(spectrum, reduction_matrix, uncertainties):
    if reduction_matrix is None:
        reduction_matrix = np.ma.ones(spectrum.shape)
        reduction_matrix.mask = np.zeros(spectrum.shape, dtype=bool)

    if isinstance(spectrum, np.ma.core.MaskedArray):
        spectral_data_corrected = np.ma.zeros(spectrum.shape)
        spectral_data_corrected.mask = copy.copy(spectrum.mask)

        if uncertainties is not None:
            pipeline_uncertainties = np.ma.masked_array(copy.copy(uncertainties))
            pipeline_uncertainties.mask = copy.copy(spectrum.mask)
        else:
            pipeline_uncertainties = None
    else:
        spectral_data_corrected = np.zeros(spectrum.shape)

        if uncertainties is not None:
            pipeline_uncertainties = copy.copy(uncertainties)
        else:
            pipeline_uncertainties = None

    return spectral_data_corrected, reduction_matrix, pipeline_uncertainties


def pipeline_validity_test(reduced_true_model, reduced_mock_observations,
                           mock_observations_reduction_matrix=None, mock_noise=None):
    if mock_observations_reduction_matrix is None:
        mock_observations_reduction_matrix = np.ones(reduced_true_model.shape)

    if mock_noise is None:
        mock_noise = np.zeros(reduced_true_model.shape)

    return 1 - (reduced_true_model - mock_noise * mock_observations_reduction_matrix) / reduced_mock_observations


def remove_noisy_wavelength_channels(spectrum, reduction_matrix, mean_subtract=False):
    for i, data in enumerate(spectrum):
        # Get standard deviation over time, for each wavelength channel
        time_standard_deviation = np.asarray([np.ma.std(data, axis=0)] * np.size(data, axis=0))

        # Mask channels where the standard deviation is greater than the total standard deviation
        data = np.ma.masked_where(
            time_standard_deviation > 3 * np.ma.std(data), data
        )

        spectrum[i, :, :] = data

    if mean_subtract:
        mean_spectra = np.mean(spectrum, axis=2)  # mean over wavelengths of each individual spectrum
        spectrum -= mean_spectra
        reduction_matrix -= mean_spectra

    return spectrum, reduction_matrix


def remove_telluric_lines_fit(spectrum, reduction_matrix, airmass, uncertainties=None, mask_threshold=1e-16,
                              polynomial_fit_degree=2, correct_uncertainties=True):
    """Remove telluric lines with a polynomial function.
    The telluric transmittance can be written as:
        T = exp(-airmass * optical_depth),
    hence the log of the transmittance can be written as a first order polynomial:
        log(T) ~ b * airmass + a.
    Using a 1st order polynomial might be not enough, as the atmospheric composition can change slowly over time. Using
    a second order polynomial, as in:
        log(T) ~ c * airmass ** 2 + b * airmass + a,
    might be safer.

    Args:
        spectrum: spectral data to correct
        reduction_matrix: matrix storing all the operations made to reduce the data
        airmass: airmass of the data
        uncertainties: uncertainties on the data
        mask_threshold: mask wavelengths where the Earth atmospheric transmittance estimate is below this value
        polynomial_fit_degree: degree of the polynomial fit of the Earth atmospheric transmittance
        correct_uncertainties:

    Returns:
        Corrected spectral data, reduction matrix and uncertainties after correction
    """
    # Initialization
    if spectrum.shape[1] <= polynomial_fit_degree + 1:
        warnings.warn(f"not enough points in airmass axis ({spectrum.shape[1]}) "
                      f"for a meaningful correction with the requested fit degree ({polynomial_fit_degree}). "
                      f"At least {polynomial_fit_degree + 2} airmass axis points are required. "
                      f"Increase the number of airmass axis points to decrease correction bias, "
                      f"or decrease the polynomial fit degree.")

    spectral_data_corrected, reduction_matrix, pipeline_uncertainties = __init_pipeline_outputs(
        spectrum, reduction_matrix, uncertainties
    )

    if uncertainties is not None:
        # The log of the spectrum is fitted, so the weights must be the weights of the log
        spectrum = np.ma.masked_equal(spectrum, 0)
        uncertainties = np.ma.masked_equal(uncertainties, 0)
        weights = 1 / np.abs(uncertainties / spectrum)  # 1 / uncertainties of the log
        weights = weights.filled(0)
    else:
        weights = np.ones(spectrum.shape)

    spectrum[np.nonzero(np.equal(weights, 0))] = 0  # ensure no invalid values are hidden where weight = 0

    telluric_lines_fits = np.ma.zeros(spectral_data_corrected.shape)

    # Correction
    for i, det in enumerate(spectrum):
        # Mask wavelength columns where at least one value is lower or equal to 0, to avoid invalid log values
        masked_det = np.ma.masked_where(np.ones(det.shape) * np.min(det, axis=0) <= 0, det)
        log_det_t = np.ma.log(np.transpose(masked_det))
        weights[i][masked_det.mask] = 0  # polyfit doesn't take masks into account, so set weight of masked values to 0

        # Fit each wavelength column
        for k, log_wavelength_column in enumerate(log_det_t):
            if np.allclose(weights[i, :, k], 0, atol=sys.float_info.min):  # skip fully masked columns
                telluric_lines_fits[i, :, k] = 0

                continue

            fit_parameters = np.polynomial.Polynomial.fit(
                x=airmass, y=log_wavelength_column, deg=polynomial_fit_degree, w=weights[i, :, k]
            )
            fit_function = np.polynomial.Polynomial(fit_parameters.convert().coef)
            telluric_lines_fits[i, :, k] = fit_function(airmass)

        # Calculate telluric transmittance estimate
        telluric_lines_fits[i, :, :] = np.exp(telluric_lines_fits[i, :, :])

        # Apply mask where estimate is lower than the threshold, as well as the data mask
        telluric_lines_fits[i, :, :] = np.ma.masked_where(
            np.ones(telluric_lines_fits[i].shape) * np.min(telluric_lines_fits[i, :, :], axis=0) < mask_threshold,
            telluric_lines_fits[i, :, :]
        )
        telluric_lines_fits[i, :, :] = np.ma.masked_where(
            masked_det.mask, telluric_lines_fits[i, :, :]
        )

        # Apply correction
        spectral_data_corrected[i, :, :] = det
        spectral_data_corrected[i, :, :] /= telluric_lines_fits[i, :, :]
        reduction_matrix[i, :, :] /= telluric_lines_fits[i, :, :]

    # Propagation of uncertainties
    if uncertainties is not None:
        pipeline_uncertainties /= np.abs(telluric_lines_fits)

        if correct_uncertainties:
            degrees_of_freedom = 1 + polynomial_fit_degree

            if np.ndim(pipeline_uncertainties.mask) != np.ndim(pipeline_uncertainties):
                raise ValueError(f"number of dimensions of the mask of pipeline uncertainties "
                                 f"({np.ndim(pipeline_uncertainties.mask)}) does not matches pipeline uncertainties"
                                 f"number of dimension ({np.ndim(pipeline_uncertainties)})")

            # Count number of non-masked points minus degrees of freedom in each time axes
            valid_points = airmass.size - np.sum(pipeline_uncertainties.mask, axis=1) - degrees_of_freedom
            valid_points[np.less(valid_points, 0)] = 0

            # Correct from fitting effect
            # Uncertainties are assumed unbiased, but fitting induces a bias, so here the uncertainties are voluntarily
            # biased (https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance)
            # This way the uncertainties truly reflect the standard deviation of the data
            pipeline_uncertainties = np.moveaxis(pipeline_uncertainties, 0, 1)
            pipeline_uncertainties *= np.sqrt(valid_points / airmass.size)
            pipeline_uncertainties = np.ma.masked_less_equal(pipeline_uncertainties, 0)
            pipeline_uncertainties = np.moveaxis(pipeline_uncertainties, 1, 0)

    return spectral_data_corrected, reduction_matrix, pipeline_uncertainties


def remove_telluric_lines_mean(spectrum, reduction_matrix, uncertainties=None, mask_threshold=1e-16):
    """Remove the telluric lines using the weighted arithmetic mean over time.

    Args:
        spectrum: spectral data to correct
        reduction_matrix: matrix storing all the operations made to reduce the data
        uncertainties: uncertainties on the data
        mask_threshold: mask wavelengths where the Earth atmospheric transmittance estimate is below this value

    Returns:
        Corrected spectral data, reduction matrix and uncertainties after correction
    """
    # Initialization
    spectral_data_corrected, reduction_matrix, pipeline_uncertainties = __init_pipeline_outputs(
        spectrum, reduction_matrix, uncertainties
    )

    if uncertainties is not None:
        weights = 1 / uncertainties
        weights[weights.mask] = 0
    else:
        weights = np.ones(spectrum.shape)

    mean_spectrum_time = np.ma.average(spectrum, axis=1, weights=weights)
    mean_spectrum_time = np.ma.masked_array(mean_spectrum_time)  # ensure that it is a masked array

    if np.ndim(mean_spectrum_time.mask) == 0:
        mean_spectrum_time.mask = np.zeros(mean_spectrum_time.shape, dtype=bool)

    if np.ndim(spectral_data_corrected.mask) == 0:
        spectral_data_corrected.mask = np.zeros(spectral_data_corrected.shape, dtype=bool)

    # Correction
    if isinstance(spectral_data_corrected, np.ma.core.MaskedArray):
        for i, data in enumerate(spectrum):
            mean_spectrum_time[i] = np.ma.masked_where(
                mean_spectrum_time[i] < mask_threshold, mean_spectrum_time[i]
            )
            spectral_data_corrected.mask[i, :, :] = mean_spectrum_time.mask[i]
            reduction_matrix.mask[i, :, :] = mean_spectrum_time.mask[i]
            spectral_data_corrected[i, :, :] = data / mean_spectrum_time[i]
            reduction_matrix[i, :, :] /= mean_spectrum_time[i]
    else:
        for i, data in enumerate(spectrum):
            spectral_data_corrected[i, :, :] = data / mean_spectrum_time[i]
            reduction_matrix[i, :, :] /= mean_spectrum_time[i]

    if uncertainties is not None:
        for i, data in enumerate(spectrum):
            pipeline_uncertainties[i, :, :] /= np.abs(mean_spectrum_time[i])

    return spectral_data_corrected, reduction_matrix, pipeline_uncertainties


def remove_throughput_fit(spectrum, reduction_matrix, wavelengths, uncertainties=None, mask_threshold=1e-16,
                          polynomial_fit_degree=2, correct_uncertainties=True):
    """Remove variable throughput with a polynomial function.

    Args:
        spectrum: spectral data to correct
        reduction_matrix: matrix storing all the operations made to reduce the data
        wavelengths: wavelengths of the data
        uncertainties: uncertainties on the data
        mask_threshold: mask wavelengths where the Earth atmospheric transmittance estimate is below this value
        polynomial_fit_degree: degree of the polynomial fit of the Earth atmospheric transmittance
        correct_uncertainties:

    Returns:
        Corrected spectral data, reduction matrix and uncertainties after correction
    """
    # Initialization
    if spectrum.shape[2] <= polynomial_fit_degree + 1:
        warnings.warn(f"not enough points in wavelengths axis ({spectrum.shape[2]}) "
                      f"for a meaningful correction with the requested fit degree ({polynomial_fit_degree}). "
                      f"At least {polynomial_fit_degree + 2} wavelengths axis points are required. "
                      f"Increase the number of wavelengths axis points to decrease correction bias, "
                      f"or decrease the polynomial fit degree.")

    spectral_data_corrected, reduction_matrix, pipeline_uncertainties = __init_pipeline_outputs(
        spectrum, reduction_matrix, uncertainties
    )

    if uncertainties is not None:
        weights = copy.deepcopy(uncertainties)  # ensure low weights within tellurics, but gives more weight to noisy non-telluric wvls
        weights = weights.filled(0)  # polyfit doesn't take masks into account, so set weight of masked values to 0
    else:
        weights = np.ones(spectrum.shape)

    spectrum[np.nonzero(np.equal(weights, 0))] = 0  # ensure no invalid values are hidden where weight = 0

    throughput_fits = np.ma.zeros(spectral_data_corrected.shape)

    if np.ndim(wavelengths) == 3:
        print('Assuming same wavelength solution for each observations, taking wavelengths of observation 0')

    # Correction
    for i, det in enumerate(spectrum):
        if np.ndim(wavelengths) == 1:
            wvl = wavelengths
        elif np.ndim(wavelengths) == 2:
            wvl = wavelengths[i, :]
        elif np.ndim(wavelengths) == 3:
            wvl = wavelengths[i, 0, :]
        else:
            raise ValueError(f"wavelengths must have at most 3 dimensions, but has {np.ndim(wavelengths)}")

        # Fit each observation
        for j, observation in enumerate(det):
            fit_parameters = np.polynomial.Polynomial.fit(
                x=wvl, y=observation, deg=polynomial_fit_degree, w=weights[i, j, :]
            )
            fit_function = np.polynomial.Polynomial(fit_parameters.convert().coef)
            throughput_fits[i, j, :] = fit_function(wvl)

        # Apply mask where estimate is lower than the threshold, as well as the data mask
        throughput_fits[i, :, :] = np.ma.masked_where(
            np.ones(throughput_fits[i].shape) * np.min(throughput_fits[i, :, :], axis=0) < mask_threshold,
            throughput_fits[i, :, :]
        )
        throughput_fits[i, :, :] = np.ma.masked_where(
            det.mask, throughput_fits[i, :, :]
        )

        # Apply correction
        spectral_data_corrected[i, :, :] = det
        spectral_data_corrected[i, :, :] /= throughput_fits[i, :, :]
        reduction_matrix[i, :, :] /= throughput_fits[i, :, :]

    # Propagation of uncertainties
    if uncertainties is not None:
        pipeline_uncertainties /= np.abs(throughput_fits)

        if correct_uncertainties:
            degrees_of_freedom = 1 + polynomial_fit_degree

            # Count number of non-masked points minus degrees of freedom in each wavelength axes
            if np.ndim(pipeline_uncertainties.mask) != np.ndim(pipeline_uncertainties):
                raise ValueError(f"number of dimensions of the mask of pipeline uncertainties "
                                 f"({np.ndim(pipeline_uncertainties.mask)}) does not matches pipeline uncertainties"
                                 f"number of dimension ({np.ndim(pipeline_uncertainties)})")

            valid_points = wavelengths.size - np.sum(pipeline_uncertainties.mask, axis=2) - degrees_of_freedom
            valid_points[np.less(valid_points, 0)] = 0

            # Correct from fitting effect
            # Uncertainties are assumed unbiased, but fitting induces a bias, so here the uncertainties are voluntarily
            # biased (https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance)
            # This way the uncertainties truly reflect the standard deviation of the data
            pipeline_uncertainties = np.moveaxis(pipeline_uncertainties, 2, 0)
            pipeline_uncertainties *= np.sqrt(valid_points / wavelengths.size)
            pipeline_uncertainties = np.ma.masked_less_equal(pipeline_uncertainties, 0)
            pipeline_uncertainties = np.moveaxis(pipeline_uncertainties, 0, 2)

    return spectral_data_corrected, reduction_matrix, pipeline_uncertainties


def remove_throughput_mean(spectrum, reduction_matrix=None, uncertainties=None):
    """Correct for the variable throughput using the weighted arithmetic mean over wavelength.

    Args:
        spectrum: spectral data to correct
        reduction_matrix: matrix storing all the operations made to reduce the data
        uncertainties: uncertainties on the data

    Returns:
        Corrected spectral data, reduction matrix and uncertainties after correction
    """
    # Initialization
    spectral_data_corrected, reduction_matrix, pipeline_uncertainties = __init_pipeline_outputs(
        spectrum, reduction_matrix, uncertainties
    )

    if uncertainties is not None:
        weights = 1 / uncertainties
        weights[weights.mask] = 0
    else:
        weights = np.ones(spectrum.shape)

    # Correction
    for i, data in enumerate(spectrum):
        if isinstance(spectrum, np.ma.core.MaskedArray):
            correction_coefficient = np.ma.average(data, axis=1, weights=weights[i])
        elif isinstance(spectrum, np.ndarray):
            correction_coefficient = np.average(data, axis=1, weights=weights[i])
        else:
            raise ValueError(f"spectral_data must be a numpy.ndarray or a numpy.ma.core.MaskedArray, "
                             f"but is of type '{type(spectrum)}'")

        spectral_data_corrected[i, :, :] = np.transpose(np.transpose(data) / correction_coefficient)
        reduction_matrix[i, :, :] = np.transpose(np.transpose(reduction_matrix[i, :, :]) / correction_coefficient)

        if uncertainties is not None:
            pipeline_uncertainties[i, :, :] = np.transpose(
                np.transpose(pipeline_uncertainties[i, :, :]) / np.abs(correction_coefficient)
            )

    return spectral_data_corrected, reduction_matrix, pipeline_uncertainties


def preparing_pipeline(spectrum, uncertainties=None,
                       wavelengths=None, airmass=None, tellurics_mask_threshold=0.1, polynomial_fit_degree=1,
                       apply_throughput_removal=True, apply_telluric_lines_removal=True, correct_uncertainties=True,
                       full=False, **kwargs):
    """Removes the telluric lines and variable throughput of some data.
    If airmass is None, the Earth atmospheric transmittance is assumed to be time-independent, so telluric transmittance
    will be fitted using the weighted arithmetic mean. Otherwise, telluric transmittance are fitted with a polynomial.

    Args:
        spectrum: spectral data to correct
        uncertainties: uncertainties on the data
        wavelengths: wavelengths of the data
        airmass: airmass of the data
        tellurics_mask_threshold: mask wavelengths where the Earth atmospheric transmittance estimate is below this value
        polynomial_fit_degree: degree of the polynomial fit of the Earth atmospheric transmittance
        apply_throughput_removal: if True, apply the throughput removal correction
        apply_telluric_lines_removal: if True, apply the telluric lines removal correction
        correct_uncertainties:
        full: if True, return the reduced matrix and reduced uncertainties in addition to the reduced spectrum

    Returns:
        Reduced spectral data (and reduction matrix and uncertainties after reduction if full is True)
    """
    # Initialize reduction matrix
    reduction_matrix = np.ma.ones(spectrum.shape)
    reduction_matrix.mask = np.zeros(spectrum.shape, dtype=bool)

    # Initialize reduced data and pipeline noise
    reduced_data = copy.copy(spectrum)

    if isinstance(spectrum, np.ma.core.MaskedArray):
        reduced_data.mask = copy.copy(spectrum.mask)

        if uncertainties is not None:
            reduced_data_uncertainties = np.ma.masked_array(copy.copy(uncertainties))
            reduced_data_uncertainties.mask = copy.copy(spectrum.mask)
        else:
            reduced_data_uncertainties = None
    else:
        if uncertainties is not None:
            reduced_data_uncertainties = copy.copy(uncertainties)
        else:
            reduced_data_uncertainties = None

    # Apply corrections
    if apply_throughput_removal:
        if wavelengths is None:
            reduced_data, reduction_matrix, reduced_data_uncertainties = remove_throughput_mean(
                spectrum=spectrum,
                reduction_matrix=reduction_matrix,
                uncertainties=reduced_data_uncertainties
            )
        else:
            reduced_data, reduction_matrix, reduced_data_uncertainties = remove_throughput_fit(
                spectrum=spectrum,
                reduction_matrix=reduction_matrix,
                wavelengths=wavelengths,
                uncertainties=reduced_data_uncertainties,
                mask_threshold=sys.float_info.min,
                polynomial_fit_degree=2,
                correct_uncertainties=correct_uncertainties
            )

    if apply_telluric_lines_removal:
        if airmass is None:
            reduced_data, reduction_matrix, reduced_data_uncertainties = remove_telluric_lines_mean(
                spectrum=reduced_data,
                reduction_matrix=reduction_matrix,
                uncertainties=reduced_data_uncertainties,
                mask_threshold=tellurics_mask_threshold
            )
        else:
            reduced_data, reduction_matrix, reduced_data_uncertainties = remove_telluric_lines_fit(
                spectrum=reduced_data,
                reduction_matrix=reduction_matrix,
                airmass=airmass,
                uncertainties=reduced_data_uncertainties,
                mask_threshold=tellurics_mask_threshold,
                polynomial_fit_degree=polynomial_fit_degree,
                correct_uncertainties=correct_uncertainties
            )

    if full:
        return reduced_data, reduction_matrix, reduced_data_uncertainties
    else:
        return reduced_data



import numpy as np
from petitRADTRANS import Radtrans

import os
os.environ["pRT_input_data_path"] = "/Users/molliere/Documents/programm_data/petitRADTRANS_public/input_data"

atmosphere = Radtrans(line_species = ['H2O_HITEMP', 'CO_all_iso_HITEMP', 'CH4', 'CO2', 'Na_allard', 'K_allard'], \
                          rayleigh_species = ['H2', 'He'], \
                          continuum_opacities = ['H2-H2', 'H2-He'], \
                          wlen_bords_micron = [0.3, 15])

atmosphere2 = Radtrans(line_species = ['H2O_HITEMP', 'CO_all_iso_Chubb', 'CH4', 'CO2', 'Na_allard', 'K_allard'], \
                          rayleigh_species = ['H2', 'He'], \
                          continuum_opacities = ['H2-H2', 'H2-He'], \
                          wlen_bords_micron = [0.3, 15])


pressures = np.logspace(-6, 2, 100)
atmosphere.setup_opa_structure(pressures)
atmosphere2.setup_opa_structure(pressures)

temperature = 1200. * np.ones_like(pressures)

abundances = {}
abundances['H2'] = 0.74 * np.ones_like(temperature)
abundances['He'] = 0.24 * np.ones_like(temperature)

abundances['H2O_HITEMP'] = 0.001 * np.ones_like(temperature)
abundances['Na_allard'] = 0.00001 * np.ones_like(temperature)
abundances['K_allard'] = 0.000001 * np.ones_like(temperature)
abundances['Na_burrows'] = 0.00001 * np.ones_like(temperature)
abundances['K_burrows'] = 0.000001 * np.ones_like(temperature)
abundances['Na_lor_cut'] = 0.00001 * np.ones_like(temperature)
abundances['K_lor_cut'] = 0.000001 * np.ones_like(temperature)

abundances['CO_all_iso_HITEMP'] = 0.01 * np.ones_like(temperature)
abundances['CO_all_iso_Chubb'] = 0.01 * np.ones_like(temperature)
abundances['CO2'] = 0.00001 * np.ones_like(temperature)
abundances['CH4'] = 0.000001 * np.ones_like(temperature)

MMW = 2.33 * np.ones_like(temperature)

from petitRADTRANS import nat_cst as nc
R_pl = 1.838*nc.r_jup_mean
gravity = 1e1**2.45
P0 = 0.01

atmosphere.calc_transm(temperature, abundances, gravity, MMW, R_pl=R_pl, P0_bar=P0)
atmosphere2.calc_transm(temperature, abundances, gravity, MMW, R_pl=R_pl, P0_bar=P0)

import pylab as plt
plt.rcParams['figure.figsize'] = (10, 6)

plt.plot(nc.c/atmosphere.freq/1e-4, atmosphere.transm_rad/nc.r_jup_mean, label = 'HITEMP')
plt.plot(nc.c/atmosphere.freq/1e-4, atmosphere2.transm_rad/nc.r_jup_mean, label = 'Exomol')

plt.xscale('log')
plt.xlabel('Wavelength (microns)')
plt.ylabel(r'Transit radius ($\rm R_{Jup}$)')
plt.legend()
plt.title('T = 1200 K')
plt.show()

temperature = 200. * np.ones_like(pressures)

atmosphere.calc_transm(temperature, abundances, gravity, MMW, R_pl=R_pl, P0_bar=P0)
atmosphere2.calc_transm(temperature, abundances, gravity, MMW, R_pl=R_pl, P0_bar=P0)

plt.plot(nc.c/atmosphere.freq/1e-4, atmosphere.transm_rad/nc.r_jup_mean, label = 'HITEMP')
plt.plot(nc.c/atmosphere.freq/1e-4, atmosphere2.transm_rad/nc.r_jup_mean, label = 'Exomol')

plt.xscale('log')
plt.xlabel('Wavelength (microns)')
plt.ylabel(r'Transit radius ($\rm R_{Jup}$)')
plt.legend()
plt.title('T = 200 K')
plt.show()



import unittest
from petitRADTRANS import Data
import petitRADTRANS.nat_cst as nc
from petitRADTRANS.retrieval.models import emission_model_diseq, isothermal_free_transmission

class TestData(unittest.TestCase):
    def test_init(self):
        self.test_fits_file = "../petitRADTRANS/retrieval/examples/emission/HR8799e_spectra.fits"        
        self.test_txt_file = "../petitRADTRANS/retrieval/examples/transmission/hst_example_clear_spec.fits"
        self.test_fits()
        self.test_txt()
        return 

    def test_fits(self):
        fits_test = Data("fits_test",
                         path_to_observations= self.test_fits_file,
                         distance = 40*nc.pc,
                         data_resolution = 500,
                         model_resolution = 500,
                         model_generating_function = emission_model_diseq,
                         scale = False)
        self.assertEqual(fits_test.path_to_observations, self.test_fits_file)
        self.assertEqual(fits_test.data_resolution,500)
        self.assertEqual(fits_test.model_resolution,500)
        self.assertEqual(fits_test.distance,40*nc.pc)
        self.assertEqual(fits_test.calculate_star_radiosity[0], 2.509834381986098e-17)
        self.assertEqual(fits_test.wlen[0],1.9700000286102295)

        fits_test.scale_to_distance(10*nc.pc)
        self.assertEqual(fits_test.distance,10*nc.pc)
        self.assertEqual(fits_test.calculate_star_radiosity[0], 2.509834381986098e-17 * (40. / 10.) ** 2)

        self.fits_test = fits_test
        return self.fits_test

    def test_txt(self):
        txt_test = Data("txt_test",
                         path_to_observations= self.test_txt_file,
                         distance = 40*nc.pc,
                         data_resolution = 50,
                         model_resolution = 50,
                         model_generating_function = isothermal_free_transmission,
                         scale = True)
        self.assertEqual(txt_test.path_to_observations, self.test_fits_file)
        self.assertEqual(txt_test.data_resolution,50)
        self.assertEqual(txt_test.model_resolution,50)
        self.assertEqual(txt_test.distance,40*nc.pc)
        self.assertEqual(txt_test.calculate_star_radiosity[0], 0.001365)
        self.assertEqual(txt_test.wlen[0],1.138000)
        self.assertEqual(txt_test.scale,True)

        txt_test.scale_to_distance(10*nc.pc)
        self.assertEqual(txt_test.distance,10*nc.pc)
        self.assertEqual(txt_test.calculate_star_radiosity[0], 0.001365 * (40. / 10.) ** 2)
        self.txt_test = txt_test
        return self.txt_test



import sys, os
import copy as cp
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
from scipy.interpolate import interp1d,CubicSpline

from petitRADTRANS import nat_cst as nc
from petitRADTRANS.retrieval import cloud_cond as fc
from petitRADTRANS import poor_mans_nonequ_chem as pm

from petitRADTRANS.physics import PT_ret_model,\
                                  guillot_global,\
                                  guillot_global_ret,\
                                  guillot_modif,\
                                  isothermal,\
                                  cubic_spline_profile,\
                                  linear_spline_profile,\
                                  dTdP_temperature_profile,\
                                  madhu_seager_2009

from .chemistry import get_abundances
from .util import surf_to_meas, calc_MMW, compute_gravity, spectrum_cgs_to_si
"""
Models Module

This module contains a set of functions that generate the spectra used
in the petitRADTRANS retrieval. This includes setting up the
pressure-temperature structure, the chemistry, and the radiative
transfer to compute the emission or transmission spectrum.

All models must take the same set of inputs:

    pRT_object : petitRADTRANS.RadTrans
        This is the pRT object that is used to compute the spectrum
        It must be fully initialized prior to be used in the model function
    parameters : dict
        A dictionary of Parameter objects. The naming of the parameters
        must be consistent between the Priors and the model function you
        are using.
    PT_plot_mode : bool
        If this argument is True, the model function should return the pressure
        and temperature arrays before computing the flux.
    AMR : bool
        If this parameter is True, your model should allow for reshaping of the
        pressure and temperature arrays based on the position of the clouds or
        the location of the photosphere, increasing the resolution where required.
        For example, using the fixed_length_amr function defined below.
"""

# Global constants to reduce calculations and initializations.
PGLOBAL = np.logspace(-6,3,1000)

def emission_model_diseq(pRT_object,
                         parameters,
                         PT_plot_mode = False,
                         AMR = True):
    """
    Disequilibrium Chemistry Emission Model

    This model computes an emission spectrum based on the temperature profile of (Molliere 2020).
    (Dis)equilibrium or free chemistry, can be used. The use of easychem for on-the-fly (dis)equilibrium
    chemistry calculations is supported, but is currently under development.
    Many of the parameters are optional, but must be used in the correct combination
    with other parameters.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  T_int : Interior temperature of the planet [K]
                *  T3 : Innermost temperature spline [K]
                *  T2 : Middle temperature spline [K]
                *  T1 : Outer temperature spline [K]
                *  alpha : power law index in tau = delta * press_cgs**alpha
                *  log_delta : proportionality factor in tau = delta * press_cgs**alpha
                *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                *  Fe/H : Metallicity
                *  C/O : Carbon to oxygen ratio
                *  fsed : sedimentation parameter - can be unique to each cloud type by adding _CloudName
                One of:
                  *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                  *  b_hans : Width of cloud particle size distribution (hansen)
                One of:
                  *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                  *  log_kzz : Vertical mixing parameter
                One of
                  *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                  *  log_X_cb_: cloud mass fraction abundance
                Optional
                  *  contribution : return the emission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR : bool
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.ndarray
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.ndarray
            Computed emission spectrum [W/m2/micron]
        contr_em : Optional, np.ndarray
            Emission contribution function, relative contributions for each wavelength and pressure level.

    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)

    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value

    # Use this for debugging.
    #for key, val in parameters.items():
    #    print(key,val.value)

    # Priors for these parameters are implemented here, as they depend on each other
    T3 = ((3./4.*parameters['T_int'].value**4.*(0.1+2./3.))**0.25)*(1.0-parameters['T3'].value)
    T2 = T3*(1.0-parameters['T2'].value)
    T1 = T2*(1.0-parameters['T1'].value)
    delta = ((10.0**(-3.0+5.0*parameters['log_delta'].value))*1e6)**(-parameters['alpha'].value)
    gravity, R_pl =  compute_gravity(parameters)

    # Make the P-T profile
    temp_arr = np.array([T1,T2,T3])
    carbon_to_oxygen = 0
    if 'C/O' in parameters.keys():
        carbon_to_oxygen = parameters['C/O'].value
    elif "C" in parameters.keys():
        carbon_to_oxygen = parameters['C'].value/parameters['O'].value
    temperatures = PT_ret_model(temp_arr, \
                            delta,
                            parameters['alpha'].value,
                            parameters['T_int'].value,
                            p_use,
                            parameters['Fe/H'].value,
                            carbon_to_oxygen,
                            conv=True)
    if 'use_easychem' in parameters.keys():
        temperatures[np.where(temperatures<40.0)] = 40.0
        temperatures[np.where(temperatures>42000.0)] = 42000.0

    # If in evaluation mode, and PTs are supposed to be plotted
    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None
    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    # Calculate the spectrum
    if pressures.shape[0] != pRT_object.press.shape[0]:
        print("Incorrect output shape!")
        return None,None

    # Hansen or log normal clouds
    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)

    # calculate the spectrum
    pRT_object.calc_flux(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        contribution = contribution,
                        fsed = fseds,
                        Kzz = kzz,
                        sigma_lnorm = sigma_lnorm,
                        b_hans = b_hans,
                        radius = radii,
                        dist = distribution)

    # Getting the model into correct units (W/m2/micron)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_em
    return wlen_model, spectrum_model

def emission_model_diseq_patchy_clouds(pRT_object,
                                       parameters,
                                       PT_plot_mode = False,
                                       AMR = True):
    """
    Disequilibrium Chemistry Emission Model

    This model computes an emission spectrum based on the temperature profile of (Molliere 2020).
    (Dis)equilibrium or free chemistry, can be used. The use of easychem for on-the-fly (dis)equilibrium
    chemistry calculations is supported, but is currently under development.
    This model includes patchy clouds, and requires a unique temperature profile for the
    clear atmosphere regions.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  T_int : Interior temperature of the planet [K]
                *  T3 : Innermost temperature spline 
                *  T2 : Middle temperature spline 
                *  T1 : Outer temperature spline 
                *  T3_clear : Innermost temperature spline for clear atmosphere
                *  T2_clear : Middle temperature spline for clear atmosphere
                *  T1_clear : Outer temperature spline for clear atmosphere
                *  alpha : power law index in tau = delta * press_cgs**alpha
                *  alpha_clear : power law index in tau = delta * press_cgs**alpha for clear atmosphere
                *  log_delta : proportionality factor in tau = delta * press_cgs**alpha
                *  log_delta_clear : proportionality factor in tau = delta * press_cgs**alpha for clear atmosphere
                *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                *  Fe/H : Metallicity
                *  C/O : Carbon to oxygen ratio
                *  fsed : sedimentation parameter - can be unique to each cloud type
                One of:
                  *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                  *  b_hans : Width of cloud particle size distribution (hansen)
                One of:
                  *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                  *  log_kzz : Vertical mixing parameter
                One of
                  *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                  *  log_X_cb_: cloud mass fraction abundance
                Optional
                  *  contribution : return the emission contribution function
                *  patchiness : Fraction of cloud coverage, clear contribution is (1-patchiness)
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR : bool
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed emission spectrum [W/m2/micron]
        contr_em : Optional, np.ndarray
            Emission contribution function, relative contributions for each wavelength and pressure level.
            Only returns the contribution of the clear atmosphere component.

    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)

    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
     # Not sure how to deal with having 2 separate contribution function

    # Priors for these parameters are implemented here, as they depend on each other
    T3 = ((3./4.*parameters['T_int'].value**4.*(0.1+2./3.))**0.25)*(1.0-parameters['T3'].value)
    T2 = T3*(1.0-parameters['T2'].value)
    T1 = T2*(1.0-parameters['T1'].value)
    delta = ((10.0**(-3.0+5.0*parameters['log_delta'].value))*1e6)**(-parameters['alpha'].value)
    temp_arr = np.array([T1,T2,T3])

    T3_clear = ((3./4.*parameters['T_int'].value**4.*(0.1+2./3.))**0.25)*(1.0-parameters['T3_clear'].value)
    T2_clear = T3_clear*(1.0-parameters['T2_clear'].value)
    T1_clear = T2_clear*(1.0-parameters['T1_clear'].value)
    temps_clear = np.array([T1_clear,T2_clear,T3_clear])
    delta_clear = ((10.0**(-3.0+5.0*parameters['log_delta_clear'].value))*1e6)**(-parameters['alpha_clear'].value)
    gravity, R_pl =  compute_gravity(parameters)

    temperatures = PT_ret_model(temp_arr,
                            delta,
                            parameters['alpha'].value,
                            parameters['T_int'].value,
                            p_use,
                            parameters['Fe/H'].value,
                            parameters['C/O'].value,
                            conv=True)

    # If in evaluation mode, and PTs are supposed to be plotted
    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)

    t_clear = PT_ret_model(temps_clear,
                            delta_clear,
                            parameters['alpha_clear'].value,
                            parameters['T_int'].value,
                            PGLOBAL[small_index],
                            parameters['Fe/H'].value,
                            parameters['C/O'].value,
                            conv=True)
    abundances_clear, MMW_clear, small_index_clear, Pbases_clear = get_abundances(PGLOBAL[small_index],
                                                  t_clear,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR = False)
    if abundances is None:
        return None, None
    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use
    if pressures.shape[0] != pRT_object.press.shape[0]:
        return None,None

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    pRT_object.calc_flux(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        contribution = contribution,
                        fsed = fseds,
                        Kzz = kzz,
                        sigma_lnorm = sigma_lnorm,
                        b_hans = b_hans,
                        radius = radii,
                        dist = distribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model_cloudy = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)

    # Set the cloud abundances to 0 for clear case
    for cloud in pRT_object.cloud_species:
        cname = cloud.split('_')[0]
        abundances_clear[cname] = np.zeros_like(temperatures)
    pRT_object.calc_flux(t_clear,
                    abundances_clear,
                    gravity,
                    MMW_clear,
                    contribution = contribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model_clear = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)

    # Patchiness fraction
    patchiness = parameters["patchiness"].value
    spectrum_model = (patchiness * spectrum_model_cloudy) +\
                     ((1-patchiness)*spectrum_model_clear)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_em
    return wlen_model, spectrum_model


def emission_model_diseq_simple_patchy_clouds(pRT_object,
                                       parameters,
                                       PT_plot_mode = False,
                                       AMR = True):
    """
    Disequilibrium Chemistry Emission Model

    This model computes an emission spectrum based on the temperature-pressure profile of (Molliere 2020).
    (Dis)equilibrium or free chemistry, can be used. The use of easychem for on-the-fly (dis)equilibrium
    chemistry calculations is supported, but is currently under development.
    This model includes patchy clouds, but uses a constant pressure temperature profile
    for both the clear and cloudy regions.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  T_int : Interior temperature of the planet [K]
                *  T3 : Innermost temperature spline [K]
                *  T2 : Middle temperature spline [K]
                *  T1 : Outer temperature spline [K]
                *  alpha : power law index in tau = delta * press_cgs**alpha
                *  log_delta : proportionality factor in tau = delta * press_cgs**alpha
                *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                *  Fe/H : Metallicity
                *  C/O : Carbon to oxygen ratio
                *  fsed : sedimentation parameter - can be unique to each cloud type
                One of:
                  *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                  *  b_hans : Width of cloud particle size distribution (hansen)
                One of:
                  *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                  *  log_kzz : Vertical mixing parameter
                One of
                  *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                  *  log_X_cb_: cloud mass fraction abundance
                *  patchiness : Fraction of cloud coverage, clear contribution is (1-patchiness)
                Optional
                  *  contribution : return the emission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed emission spectrum [W/m2/micron]
        contr_em : Optional, np.ndarray
            Emission contribution function, relative contributions for each wavelength and pressure level.
            Only returns the contribution of the clear atmosphere component.
    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)

    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
     # Not sure how to deal with having 2 separate contribution function

    # Priors for these parameters are implemented here, as they depend on each other
    T3 = ((3./4.*parameters['T_int'].value**4.*(0.1+2./3.))**0.25)*(1.0-parameters['T3'].value)
    T2 = T3*(1.0-parameters['T2'].value)
    T1 = T2*(1.0-parameters['T1'].value)
    delta = ((10.0**(-3.0+5.0*parameters['log_delta'].value))*1e6)**(-parameters['alpha'].value)
    temp_arr = np.array([T1,T2,T3])

    gravity, R_pl =  compute_gravity(parameters)

    temperatures = PT_ret_model(temp_arr,
                            delta,
                            parameters['alpha'].value,
                            parameters['T_int'].value,
                            p_use,
                            parameters['Fe/H'].value,
                            parameters['C/O'].value,
                            conv=True)

    # If in evaluation mode, and PTs are supposed to be plotted
    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        if contribution:
            return None,None,None
        return None, None
    
    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use
    if pressures.shape[0] != pRT_object.press.shape[0]:
        if contribution:
            return None,None,None
        return None, None

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    pRT_object.calc_flux(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        contribution = contribution,
                        fsed = fseds,
                        Kzz = kzz,
                        sigma_lnorm = sigma_lnorm,
                        b_hans = b_hans,
                        radius = radii,
                        dist = distribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model_cloudy = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)

    # Set the cloud abundances to 0 for clear case
    for cloud in pRT_object.cloud_species:
        cname = cloud.split('_')[0]
        abundances[cname] = np.zeros_like(temperatures)
    pRT_object.calc_flux(temperatures,
                    abundances,
                    gravity,
                    MMW,
                    contribution = contribution,
                    dist = distribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model_clear = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)

    # Patchiness fraction
    patchiness = parameters["patchiness"].value
    spectrum_model = (patchiness * spectrum_model_cloudy) +\
                     ((1-patchiness)*spectrum_model_clear)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_em  
    return wlen_model, spectrum_model

def guillot_emission(pRT_object, \
                     parameters, \
                     PT_plot_mode = False,
                     AMR = False):
    """
    Emission spectrum calculation for the Guillot 2010 temperature profile.
    (Dis)equilibrium or free chemistry, can be used. The use of easychem for on-the-fly (dis)equilibrium
    chemistry calculations is supported, but is currently under development.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  T_int : Interior temperature of the planet [K]
                *  T_equ : Equilibrium temperature of the planet
                *  gamma : Guillot gamma parameter
                *  log_kappa_IR : The log of the ratio between the infrared and optical opacities

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species

                *  fsed : sedimentation parameter - can be unique to each cloud type
                One of:
                  *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                  *  b_hans : Width of cloud particle size distribution (hansen)
                One of:
                  *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                  *  log_kzz : Vertical mixing parameter
                One of
                  *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                  *  log_X_cb_: cloud mass fraction abundance
                Optional
                  *  contribution : return the emission contribution function
        PT_plot_mode : bool
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, Emission contribution function, relative contributions for each wavelength and pressure level.

    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)

    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    gravity, R_pl = compute_gravity(parameters)

    temperatures = guillot_global(p_use, \
                                10**parameters['log_kappa_IR'].value,
                                parameters['gamma'].value, \
                                gravity, \
                                parameters['T_int'].value, \
                                parameters['T_equ'].value)

    # If in evaluation mode, and PTs are supposed to be plotted
    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None

    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    pRT_object.calc_flux(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        contribution = contribution,
                        fsed = fseds,
                        Kzz = kzz,
                        sigma_lnorm = sigma_lnorm,
                        b_hans = b_hans,
                        radius = radii,
                        dist = distribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_em
    return wlen_model, spectrum_model

def guillot_patchy_emission(pRT_object, \
                     parameters, \
                     PT_plot_mode = False,
                     AMR = False):
    """
    This model computes a emission spectrum based the Guillot temperature-pressure profile and patchy clouds.
    Either free or equilibrium chemistry can be used, together with a range of cloud parameterizations.
    It is possible to use free abundances for some species and equilibrium chemistry for the remainder.
    
    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  T_int : Interior temperature of the planet [K]
                *  T_equ : Equilibrium temperature of the planet
                *  gamma : Guillot gamma parameter
                *  log_kappa_IR : The log of the ratio between the infrared and optical opacities

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species

                *  fsed : sedimentation parameter - can be unique to each cloud type
                One of:
                  *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                  *  b_hans : Width of cloud particle size distribution (hansen)
                One of:
                  *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                  *  log_kzz : Vertical mixing parameter
                One of
                  *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                  *  log_X_cb_: cloud mass fraction abundance
                *  patchiness : Fraction of cloud coverage, clear contribution is (1-patchiness)
                Optional
                  *  contribution : return the emission contribution function
        PT_plot_mode : bool
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, the emission contribution function, relative contributions for each wavelength and pressure level.
            Only returns the contribution of the clear atmosphere component.

    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)

    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    gravity, R_pl = compute_gravity(parameters)

    temperatures = guillot_global(p_use, \
                                10**parameters['log_kappa_IR'].value,
                                parameters['gamma'].value, \
                                gravity, \
                                parameters['T_int'].value, \
                                parameters['T_equ'].value)

    # If in evaluation mode, and PTs are supposed to be plotted
    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None

    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    pRT_object.calc_flux(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        contribution = contribution,
                        fsed = fseds,
                        Kzz = kzz,
                        sigma_lnorm = sigma_lnorm,
                        b_hans = b_hans,
                        radius = radii,
                        dist = distribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model_cloudy = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)

    # Set the cloud abundances to 0 for clear case
    for cloud in pRT_object.cloud_species:
        cname = cloud.split('_')[0]
        abundances[cname] = np.zeros_like(temperatures)
    pRT_object.calc_flux(temperatures,
                    abundances,
                    gravity,
                    MMW,
                    contribution = contribution,
                    dist = distribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model_clear = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)

    # Patchiness fraction
    patchiness = parameters["patchiness"].value
    spectrum_model = (patchiness * spectrum_model_cloudy) +\
                     ((1-patchiness)*spectrum_model_clear)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_em
    return wlen_model, spectrum_model   

def interpolated_profile_emission(pRT_object, \
                                  parameters, \
                                  PT_plot_mode = False,
                                  AMR = False):
    """
    This model computes a emission spectrum based a spline temperature-pressure profile.
    Either free or equilibrium chemistry can be used, together with a range of cloud parameterizations.
    It is possible to use free abundances for some species and equilibrium chemistry for the remainder.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  nnodes : number of nodes to interplate, excluding the first and last points.
                            so the total number of nodes is nnodes + 2
                *  Temps : One parameter for each temperature node
                *  gamma : weight for penalizing the profile curvature

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species
                Optional:
                *  fsed : sedimentation parameter - can be unique to each cloud type
                One of:
                  *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                  *  b_hans : Width of cloud particle size distribution (hansen)
                One of:
                  *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                  *  log_kzz : Vertical mixing parameter
                One of
                  *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                  *  log_X_cb_: cloud mass fraction abundance
                Optional
                  *  contribution : return the emission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, the emission contribution function, relative contributions for each wavelength and pressure level.
    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)

    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    gravity, R_pl = compute_gravity(parameters)

    temp_arr = np.array([parameters[f"T{i}"].value for i in range(parameters['nnodes'].value + 2)])
    if "linear" in parameters.keys():
        temperatures, log_prior_weight = linear_spline_profile(p_use,
                                             temp_arr,
                                             parameters['gamma'].value,
                                             parameters['nnodes'].value)
    else:
        temperatures, log_prior_weight = cubic_spline_profile(p_use,
                                        temp_arr,
                                        parameters['gamma'].value,
                                        parameters['nnodes'].value)
    parameters['log_prior_weight'].value = log_prior_weight
    # If in evaluation mode, and PTs are supposed to be plotted
    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None

    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    pRT_object.calc_flux(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        contribution = contribution,
                        fsed = fseds,
                        Kzz = kzz,
                        sigma_lnorm = sigma_lnorm,
                        b_hans = b_hans,
                        radius = radii,
                        dist = distribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_em
    return wlen_model, spectrum_model

def gradient_profile_emission(pRT_object, \
                                  parameters, \
                                  PT_plot_mode = False,
                                  AMR = False):
    """
    This model computes a emission spectrum based a gradient temperature-pressure profile (Zhang 2023).
    Either free or equilibrium chemistry can be used, together with a range of cloud parameterizations.
    It is possible to use free abundances for some species and equilibrium chemistry for the remainder.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  N_layers : number of nodes to interplate, excluding the first and last points.
                            so the total number of nodes is nnodes + 2
                *  T_bottom : Temperature at the base of the atmosphere
                *  PTslope_* : Temperature gradient for each of the n_layers between which the profile is interpolated.

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species
                Optional:
                *  fsed : sedimentation parameter - can be unique to each cloud type
                One of:
                  *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                  *  b_hans : Width of cloud particle size distribution (hansen)
                One of:
                  *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                  *  log_kzz : Vertical mixing parameter
                One of
                  *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                  *  log_X_cb_: cloud mass fraction abundance
                Optional
                  *  contribution : return the emission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, the emission contribution function, relative contributions for each wavelength and pressure level.
    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)

    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    gravity, R_pl = compute_gravity(parameters)

    num_layer = parameters['N_layers'].value
    ## 1.4 assemble the P-T slopes for these layers
    layer_pt_slopes = np.ones(num_layer) * np.nan
    for index in range(num_layer):
        layer_pt_slopes[index] = parameters[f'PTslope_{num_layer - index}'].value

    temperatures = dTdP_temperature_profile(p_use,
                                            num_layer,
                                            layer_pt_slopes,
                                            parameters['T_bottom'].value)

    # If in evaluation mode, and PTs are supposed to be plotted
    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None

    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    pRT_object.calc_flux(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        contribution = contribution,
                        fsed = fseds,
                        Kzz = kzz,
                        sigma_lnorm = sigma_lnorm,
                        b_hans = b_hans,
                        radius = radii,
                        dist = distribution)
    wlen_model, f_lambda = spectrum_cgs_to_si(pRT_object.freq, pRT_object.flux)
    spectrum_model = surf_to_meas(f_lambda,
                                  R_pl,
                                  parameters['D_pl'].value)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_em
    return wlen_model, spectrum_model

def guillot_transmission(pRT_object, \
                         parameters, \
                         PT_plot_mode = False,
                         AMR = False):
    """
    Transmission Model, Guillot Profile

    This model computes a transmission spectrum based on the Guillot profile
    Either free or (dis)equilibrium chemistry can be used, together with a range of cloud parameterizations.
    It is possible to use free abundances for some species and equilibrium chemistry for the remainder.
    Chemical clouds can be used, or a simple gray opacity source.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  T_int : Interior temperature of the planet [K]
                *  T_equ : Equilibrium temperature of the planet
                *  gamma : Guillot gamma parameter
                *  log_kappa_IR : The log of the ratio between the infrared and optical opacities

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species

                Either:
                  * [log_]Pcloud : The (log) pressure at which to place the gray cloud opacity.
                Or:
                  *  fsed : sedimentation parameter - can be unique to each cloud type
                  One of:
                    *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                    *  b_hans : Width of cloud particle size distribution (hansen)
                  One of:
                    *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                    *  log_kzz : Vertical mixing parameter
                  One of
                    *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                    *  log_X_cb_: cloud mass fraction abundance
                Optional
                  *  contribution : return the transmission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, the transmission contribution function, relative contributions for each wavelength and pressure level.
    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)
    p_reference = 100.0
    if "reference_pressure" in parameters.keys():
        p_reference = parameters["reference_pressure"].value
    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    # Calculate the spectrum
    gravity, R_pl =  compute_gravity(parameters)

    temperatures = guillot_global(p_use, \
                                    10**parameters['log_kappa_IR'].value, \
                                    parameters['gamma'].value, \
                                    gravity, \
                                    parameters['T_int'].value, \
                                    parameters['T_equ'].value)


    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None

    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    pcloud = None
    gamma_scat = None
    kappa_0 = None
    if 'log_Pcloud' in parameters.keys():
        pcloud = 10**parameters['log_Pcloud'].value
    elif 'Pcloud' in parameters.keys():
        pcloud = parameters['Pcloud'].value
    if "gamma_scat" in parameters.keys():
        gamma_scat = parameters["gamma_scat"].value
    if "kappa_0" in parameters.keys():
        kappa_0 = 10**parameters["kappa_0"].value
    # Calculate the spectrum
    if len(pRT_object.cloud_species)> 0:
        sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
        pRT_object.calc_transm(temperatures,
                                abundances,
                                gravity,
                                MMW,
                                R_pl=R_pl,
                                P0_bar=p_reference,
                                sigma_lnorm = sigma_lnorm,
                                radius = radii,
                                fsed = fseds,
                                Kzz = kzz,
                                kappa_zero = kappa_0,
                                gamma_scat = gamma_scat,
                                b_hans = b_hans,
                                dist = distribution,
                                contribution = contribution)
    elif pcloud is not None:
        pRT_object.calc_transm(temperatures, \
                        abundances, \
                        gravity, \
                        MMW, \
                        R_pl=R_pl, \
                        P0_bar=p_reference,
                        Pcloud = pcloud,
                        kappa_zero = kappa_0,
                        gamma_scat = gamma_scat,
                        contribution = contribution)
    else:
        pRT_object.calc_transm(temperatures, \
                               abundances, \
                               gravity, \
                               MMW, \
                               R_pl=R_pl, \
                               P0_bar=p_reference,
                               contribution = contribution)

    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_tr
    return wlen_model, spectrum_model

def guillot_patchy_transmission(pRT_object, \
                                    parameters, \
                                    PT_plot_mode = False,
                                    AMR = False):
    """
    Transmission Model, Guillot Profile

    This model computes a transmission spectrum based on a Guillot temperature-pressure profile.
    Either free or equilibrium chemistry can be used, together with a range of cloud parameterizations.
    It is possible to use free abundances for some species and equilibrium chemistry for the remainder.
    Chemical clouds can be used, or a simple gray opacity source. This model requires patchy clouds.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  T_int : Interior temperature of the planet [K]
                *  T_equ : Equilibrium temperature of the planet
                *  gamma : Guillot gamma parameter
                *  log_kappa_IR : The log of the ratio between the infrared and optical opacities

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species

                Either:
                  * [log_]Pcloud : The (log) pressure at which to place the gray cloud opacity.
                Or:
                  *  fsed : sedimentation parameter - can be unique to each cloud type
                  One of:
                    *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                    *  b_hans : Width of cloud particle size distribution (hansen)
                  One of:
                    *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                    *  log_kzz : Vertical mixing parameter
                  One of
                    *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                    *  log_X_cb_: cloud mass fraction abundance
                *  patchiness : Fraction of cloud coverage, clear contribution is (1-patchiness)
                Optional
                  *  contribution : return the transmission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, the transmission contribution function, relative contributions for each wavelength and pressure level.
            Only the clear atmosphere contribution is returned.
    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)
    p_reference = 100.0
    if "reference_pressure" in parameters.keys():
        p_reference = parameters["reference_pressure"].value
    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    # Calculate the spectrum
    gravity, R_pl =  compute_gravity(parameters)

    temperatures = guillot_global(p_use, \
                                    10**parameters['log_kappa_IR'].value, \
                                    parameters['gamma'].value, \
                                    gravity, \
                                    parameters['T_int'].value, \
                                    parameters['T_equ'].value)

    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None
    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    # Hazes
    gamma_scat = None
    kappa_0 = None
    if "gamma_scat" in parameters.keys():
        gamma_scat = parameters["gamma_scat"].value
    if "kappa_0" in parameters.keys():
        kappa_0 = 10**parameters["kappa_0"].value

    # Calc cloudy spectrum
    pRT_object.calc_transm(temperatures,
                                abundances,
                                gravity,
                                MMW,
                                R_pl=R_pl,
                                P0_bar=p_reference,
                                sigma_lnorm = sigma_lnorm,
                                radius = radii,
                                fsed = fseds,
                                Kzz = kzz,
                                kappa_zero = kappa_0,
                                gamma_scat = gamma_scat,
                                b_hans = b_hans,
                                dist = distribution,
                                contribution = contribution)

    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model_cloudy = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    for cloud in pRT_object.cloud_species:
        cname = cloud.split('_')[0]
        abundances[cname] = np.zeros_like(temperatures)
    pRT_object.calc_transm(temperatures, \
                            abundances, \
                            gravity, \
                            MMW, \
                            R_pl=R_pl, \
                            P0_bar=p_reference,
                            contribution = contribution)

    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model_clear = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    patchiness = parameters["patchiness"].value
    spectrum_model = (patchiness * spectrum_model_cloudy) +\
                     ((1-patchiness)*spectrum_model_clear)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_tr
    return wlen_model, spectrum_model

def madhu_seager_patchy_transmission(pRT_object, \
                                     parameters, \
                                     PT_plot_mode = False,
                                     AMR = False):
    """
    Transmission Model, Madhusudhan Seager 2009 Profile

    This model computes a transmission spectrum based on a Guillot temperature-pressure profile.
    Either free or equilibrium chemistry can be used, together with a range of cloud parameterizations.
    It is possible to use free abundances for some species and equilibrium chemistry for the remainder.
    Chemical clouds can be used, or a simple gray opacity source. This model requires patchy clouds.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  log_P_set : Pressure value to contrain the PT profile, defaults to 10 bar.
                *  T_set : temperature at P_set to constrain the PT profile. [K]
                *  log_P3 : (log) Pressure value for the top of the deep atmospheric layer, [bar]
                *  P2 : in range (0,1), sets the pressure level of the middle atmospheric layer
                *  P1 : in range (0,1), sets the pressure level of the top atmospheric layer
                *  alpha_0 : slope of the upper atmospheric layer
                *  alpha_1 : slope of the middle atmospheric layer
                Optional : 
                    *  beta : power law for the slopes, default value is 0.5. Not recommended to change!

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species

                Either:
                  * [log_]Pcloud : The (log) pressure at which to place the gray cloud opacity.
                Or:
                  *  fsed : sedimentation parameter - can be unique to each cloud type
                  One of:
                    *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                    *  b_hans : Width of cloud particle size distribution (hansen)
                  One of:
                    *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                    *  log_kzz : Vertical mixing parameter
                  One of
                    *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                    *  log_X_cb_: cloud mass fraction abundance
                *  patchiness : Fraction of cloud coverage, clear contribution is (1-patchiness)
                Optional
                  *  contribution : return the transmission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, the transmission contribution function, relative contributions for each wavelength and pressure level.
            Only the clear atmosphere contribution is returned.
    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)
    p_reference = 100.0
    if "reference_pressure" in parameters.keys():
        p_reference = parameters["reference_pressure"].value
    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    # Calculate the spectrum
    gravity, R_pl =  compute_gravity(parameters)

    # Set up pressure points, guaranteeing P3>P2>P1 >= P_top
    offset = np.log10(p_use[0])
    logP3 = parameters['log_P3'].value - offset
    logP2 = logP3*(1.0-parameters['P2'].value)
    logP1 = logP2*(1.0-parameters['P1'].value)
    pressure_points = [p_use[0],(logP1 + offset),(logP2 + offset),(logP3 + offset), parameters["log_P_set"].value]

    alpha_points = [parameters["alpha_0"].value, parameters["alpha_1"].value]
    beta_points = [0.5,0.5]
    if "beta" in parameters.keys():
        beta_points = [parameters["beta"].value, parameters["beta"].value]

    temperatures = madhu_seager_2009(p_use,
                                    pressure_points,
                                    parameters["T_set"].value,
                                    alpha_points,
                                    beta_points)

    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None
    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    # Hazes
    gamma_scat = None
    kappa_0 = None
    if "gamma_scat" in parameters.keys():
        gamma_scat = parameters["gamma_scat"].value
    if "kappa_0" in parameters.keys():
        kappa_0 = 10**parameters["kappa_0"].value

    # Calc cloudy spectrum
    pRT_object.calc_transm(temperatures,
                                abundances,
                                gravity,
                                MMW,
                                R_pl=R_pl,
                                P0_bar=p_reference,
                                sigma_lnorm = sigma_lnorm,
                                radius = radii,
                                fsed = fseds,
                                Kzz = kzz,
                                kappa_zero = kappa_0,
                                gamma_scat = gamma_scat,
                                b_hans = b_hans,
                                dist = distribution,
                                contribution = contribution)

    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model_cloudy = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    if "patchiness" in parameters.key():
        for cloud in pRT_object.cloud_species:
            cname = cloud.split('_')[0]
            abundances[cname] = np.zeros_like(temperatures)
        pRT_object.calc_transm(temperatures, \
                                abundances, \
                                gravity, \
                                MMW, \
                                R_pl=R_pl, \
                                P0_bar=p_reference,
                                contribution = contribution)

        wlen_model = nc.c/pRT_object.freq/1e-4
        spectrum_model_clear = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
        patchiness = parameters["patchiness"].value
        spectrum_model = (patchiness * spectrum_model_cloudy) +\
                        ((1-patchiness)*spectrum_model_clear)
    else:
        spectrum_model = spectrum_model_cloudy

    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_tr
    return wlen_model, spectrum_model

def guillot_patchy_transmission_constrained_chem(pRT_object, \
                                                parameters, \
                                                PT_plot_mode = False,
                                                AMR = False):
    """
    Transmission Model, Guillot Profile

    This model computes a transmission spectrum based on a Guillot temperature-pressure profile.
    Either free or equilibrium chemistry can be used, together with a range of cloud parameterizations.
    It is possible to use free abundances for some species and equilibrium chemistry for the remainder.
    Chemical clouds can be used, or a simple gray opacity source. This model requires patchy clouds.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  T_int : Interior temperature of the planet [K]
                *  T_equ : Equilibrium temperature of the planet
                *  gamma : Guillot gamma parameter
                *  log_kappa_IR : The log of the ratio between the infrared and optical opacities

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species

                Either:
                  * [log_]Pcloud : The (log) pressure at which to place the gray cloud opacity.
                Or:
                  *  fsed : sedimentation parameter - can be unique to each cloud type
                  One of:
                    *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                    *  b_hans : Width of cloud particle size distribution (hansen)
                  One of:
                    *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                    *  log_kzz : Vertical mixing parameter
                  One of
                    *  eq_scaling_* : Scaling factor for equilibrium cloud abundances.
                    *  log_X_cb_: cloud mass fraction abundance
                *  patchiness : Fraction of cloud coverage, clear contribution is (1-patchiness)
                Optional
                  *  contribution : return the transmission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, the transmission contribution function, relative contributions for each wavelength and pressure level.
            Only the clear atmosphere contribution is returned.
    """
    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)
    p_reference = 100.0
    if "reference_pressure" in parameters.keys():
        p_reference = parameters["reference_pressure"].value
    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    # Calculate the spectrum
    gravity, R_pl =  compute_gravity(parameters)
    if parameters["H2O_Exomol"].value < parameters["CO2"].value: return None,None
    if parameters["CO_all_iso_HITEMP"].value < parameters["CO2"].value: return None,None

    temperatures = guillot_global(p_use, \
                                    10**parameters['log_kappa_IR'].value, \
                                    parameters['gamma'].value, \
                                    gravity, \
                                    parameters['T_int'].value, \
                                    parameters['T_equ'].value)

    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)

    if abundances is None:
        return None, None
    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
    # Hazes
    gamma_scat = None
    kappa_0 = None
    if "gamma_scat" in parameters.keys():
        gamma_scat = parameters["gamma_scat"].value
    if "kappa_0" in parameters.keys():
        kappa_0 = 10**parameters["kappa_0"].value

    # Calc cloudy spectrum
    pRT_object.calc_transm(temperatures,
                                abundances,
                                gravity,
                                MMW,
                                R_pl=R_pl,
                                P0_bar=p_reference,
                                sigma_lnorm = sigma_lnorm,
                                radius = radii,
                                fsed = fseds,
                                Kzz = kzz,
                                kappa_zero = kappa_0,
                                gamma_scat = gamma_scat,
                                b_hans = b_hans,
                                dist = distribution,
                                contribution = contribution)

    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model_cloudy = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    for cloud in pRT_object.cloud_species:
        cname = cloud.split('_')[0]
        abundances[cname] = np.zeros_like(temperatures)
    pRT_object.calc_transm(temperatures, \
                            abundances, \
                            gravity, \
                            MMW, \
                            R_pl=R_pl, \
                            P0_bar=p_reference,
                            contribution = contribution)

    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model_clear = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    patchiness = parameters["patchiness"].value
    spectrum_model = (patchiness * spectrum_model_cloudy) +\
                     ((1-patchiness)*spectrum_model_clear)
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_tr
    return wlen_model, spectrum_model

def isothermal_transmission(pRT_object, \
                            parameters, \
                            PT_plot_mode = False,
                            AMR = False):
    """
    Equilibrium Chemistry Transmission Model, Isothermal Profile

    This model computes a transmission spectrum based on an isothermal temperature-pressure profile.

    Args:
        pRT_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                *  D_pl : Distance to the planet in [cm]
                Two of
                  *  log_g : Log of surface gravity
                  *  R_pl : planet radius [cm]
                  *  mass : planet mass [g]
                *  Temp : Interior temperature of the planet [K]

                Either:
                  *  log_pquench : Pressure at which CO, CH4 and H2O abundances become vertically constant
                  *  Fe/H : Metallicity
                  *  C/O : Carbon to oxygen ratio
                Or:
                  * $SPECIESNAME[_$DATABASE][_R_$RESOLUTION] : The log mass fraction abundance of the species

                Either:
                  * [log_]Pcloud : The (log) pressure at which to place the gray cloud opacity.
                Or:
                  *  fsed : sedimentation parameter - can be unique to each cloud type
                  One of:
                    *  sigma_lnorm : Width of cloud particle size distribution (log normal)
                    *  b_hans : Width of cloud particle size distribution (hansen)
                  One of:
                    *  log_cloud_radius_* : Central particle radius (typically computed with fsed and Kzz)
                    *  log_kzz : Vertical mixing parameter
                Optional
                  *  patchiness : Fraction of cloud coverage, clear contribution is (1-patchiness)
                  *  contribution : return the transmission contribution function
        PT_plot_mode : bool
            Return only the pressure-temperature profile for plotting. Evaluate mode only.
        AMR :
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base.

    Returns:
        wlen_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
        contr-em : np.ndarray
            Optional, the transmission contribution function, relative contributions for each wavelength and pressure level.
            Only the clear atmosphere contribution is returned if patchy clouds are considered.
    """

    p_use = initialize_pressure(pRT_object.press/1e6, parameters, AMR)
    p_reference = 100.0
    if "reference_pressure" in parameters.keys():
        p_reference = parameters["reference_pressure"].value
    # Make the P-T profile
    temperatures = isothermal(p_use,parameters["Temp"].value)
    gravity, R_pl = compute_gravity(parameters)

    contribution = False
    if "contribution" in parameters.keys():
        contribution = parameters["contribution"].value
    # Make the abundance profile
    abundances, MMW, small_index, Pbases = get_abundances(p_use,
                                                  temperatures,
                                                  pRT_object.line_species,
                                                  pRT_object.cloud_species,
                                                  parameters,
                                                  AMR =AMR)
    if abundances is None:
        return None, None
    if PT_plot_mode:
        return p_use[small_index], temperatures[small_index]
    if AMR:
        temperatures = temperatures[small_index]
        pressures = PGLOBAL[small_index]
        MMW = MMW[small_index]
        pRT_object.press = pressures * 1e6
    else:
        pressures = p_use

    # Calculate the spectrum
    pcloud = None
    kappa_0 = None
    gamma_scat = None
    if 'log_Pcloud' in parameters.keys():
        pcloud = 10**parameters['log_Pcloud'].value
    elif 'Pcloud' in parameters.keys():
        pcloud = parameters['Pcloud'].value
    if "gamma_scat" in parameters.keys():
        gamma_scat = parameters["gamma_scat"].value
    if "kappa_0" in parameters.keys():
        kappa_0 = 10**parameters["kappa_0"].value
    # Calculate the spectrum
    if len(pRT_object.cloud_species)> 0:
        sigma_lnorm, fseds, kzz, b_hans, radii, distribution = fc.setup_clouds(pressures, parameters, pRT_object.cloud_species)
        pRT_object.calc_transm(temperatures,
                                abundances,
                                gravity,
                                MMW,
                                R_pl=R_pl,
                                P0_bar=p_reference,
                                sigma_lnorm = sigma_lnorm,
                                radius = radii,
                                fsed = fseds,
                                Kzz = kzz,
                                kappa_zero = kappa_0,
                                gamma_scat = gamma_scat,
                                b_hans = b_hans,
                                dist = distribution,
                                contribution = contribution)
    elif pcloud is not None:
        pRT_object.calc_transm(temperatures,
                        abundances,
                        gravity,
                        MMW,
                        R_pl=R_pl,
                        P0_bar=p_reference,
                        Pcloud = pcloud,
                        kappa_zero = kappa_0,
                        gamma_scat = gamma_scat,
                        contribution = contribution)
    else:
        pRT_object.calc_transm(temperatures,
                               abundances,
                               gravity,
                               MMW,
                               R_pl=R_pl,
                               P0_bar=p_reference,
                               kappa_zero = kappa_0,
                               gamma_scat = gamma_scat,
                               contribution = contribution)
    wlen_model = nc.c/pRT_object.freq/1e-4
    spectrum_model = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
    if contribution:
        return wlen_model, spectrum_model, pRT_object.contr_tr
    if "patchiness" in parameters.keys():
        if len(pRT_object.cloud_species) > 0 :
            for cloud in pRT_object.cloud_species:
                abundances[cloud.split('_')[0]] = np.zeros_like(temperatures)
        pRT_object.calc_transm(temperatures, \
                        abundances, \
                        gravity, \
                        MMW, \
                        R_pl=R_pl, \
                        P0_bar=p_reference,
                        Pcloud = None,
                        contribution = contribution)
        wlen_model_clear = nc.c/pRT_object.freq/1e-4
        spectrum_model_clear = (pRT_object.transm_rad/parameters['Rstar'].value)**2.
        patchiness = parameters["patchiness"].value
        spectrum_model_full = (patchiness * spectrum_model) +\
                            ((1-patchiness)*spectrum_model_clear)
        return wlen_model, spectrum_model_full
    return wlen_model, spectrum_model

def initialize_pressure(press, parameters, AMR):
    """
    Provide the pressure array correctly sized to the pRT_object in use, accounting for
    the use of Adaptive Mesh Refinement around the location of clouds.

    Args:
        press : numpy.ndarray
            Pressure array from a pRT_object. Used to set the min and max values of PGLOBAL
        shape : int
            the shape of the pressure array if no AMR is used
        scaling :
            The factor by which the pressure array resolution should be scaled.
    """
    if AMR:
        set_pglobal(press, parameters)
        p_use = PGLOBAL
    else:
        p_use = press
    return p_use

def set_pglobal(press, parameters):
    """
    Check to ensure that the global pressure array has the correct length.
    Updates PGLOBAL.

    Args:
        press : numpy.ndarray
            Pressure array from a pRT_object. Used to set the min and max values of PGLOBAL
        parameters : dict
            Must include the 'pressure_simple' and 'pressure_scaling' parameters,
            used to determine the size of the high resolution grid.
    """
    try:
        pglobal_check(press,
                    parameters['pressure_simple'].value,
                    parameters['pressure_scaling'].value)
    except KeyError():
        print("You must include the pressure_simple and pressure_scaling parameters when using AMR!")
        sys.exit(1)

def pglobal_check(press,shape,scaling):
    """
    Check to ensure that the global pressure array has the correct length.
    Updates PGLOBAL.

    Args:
        press : numpy.ndarray
            Pressure array from a pRT_object. Used to set the min and max values of PGLOBAL
        shape : int
            the shape of the pressure array if no AMR is used
        scaling :
            The factor by which the pressure array resolution should be scaled.
    """
    global PGLOBAL
    if PGLOBAL.shape[0] != int(scaling*shape):
        PGLOBAL = np.logspace(np.log10(press[0]),
                              np.log10(press[-1]),
                              int(scaling*shape))





import os
os.environ["OMP_NUM_THREADS"] = "1"
import copy as cp
import numpy as np
try:
    import easychem as ec
except ModuleNotFoundError:
    pass
from petitRADTRANS import poor_mans_nonequ_chem as pm
from petitRADTRANS.retrieval import cloud_cond as fc
from petitRADTRANS.retrieval.util import fixed_length_amr, calc_MMW

def get_abundances(pressures, temperatures, line_species, cloud_species, parameters, AMR = False):
    """
    This function takes in the C/O ratio, metallicity, and quench pressures and uses them
    to compute the gas phase and equilibrium condensate abundances from an interpolated table.
    Alternatively, a "free chemistry" approach can be used to set abundances.
    This function assumes a hydrogen-helium dominated atmosphere, and enforces <100% trace gas
    abundance by mass.

    Args:
        pressures : numpy.ndarray
            A log spaced pressure array. If AMR is on it should be the full high resolution grid.
        temperatures : numpy.ndarray
            A temperature array with the same shape as pressures
        line_species : List(str)
            A list of gas species that will contribute to the line-by-line opacity of the pRT atmosphere.
        cloud_species : List(str)
            A list of condensate species that will contribute to the cloud opacity of the pRT atmosphere.
        parameters : dict
            A dictionary of model parameters, in particular it must contain the names C/O, Fe/H and
            log_pquench. Additionally the cloud parameters log_X_cb_Fe(c) and MgSiO3(c) must be present.
        AMR : bool
            Turn the adaptive mesh grid on or off. See fixed_length_amr for implementation.

    Returns:
        abundances : dict
            Mass fraction abundances of all atmospheric species
        MMW : numpy.ndarray
            Array of the mean molecular weights in each pressure bin
        small_index : numpy.ndarray
            The indices of the high resolution grid to use to define the adaptive grid.
        PBases : dict
            A dictionary of the cloud base pressures, either computed from equilibrium
            condensation or set by the user.
    """
    # Free Chemistry
    abundances_interp = {}
    if 'use_easychem' in parameters.keys():
        # Actual equilibrium chemistry
        # Can retrieve atomic abundances

        # Calling it abundances_interp to be consistent w poor mans
        """abundances_interp = get_easychem_abundances(pressures,
                                                    temperatures,
                                                    line_species,
                                                    cloud_species,
                                                    parameters)"""
        abundances_interp = get_exoatmos_abundances(pressures,
                                                    temperatures,
                                                    parameters)
        MMW = abundances_interp['MMW']
    elif "C/O" in parameters.keys():
        # Check C/O AFTER easychem check -> need to use poor mans

        # Interpolated Equilibrium chemistry
        # Make the abundance profile
        pquench_C = None
        if 'log_pquench' in parameters.keys():
            pquench_C = 10**parameters['log_pquench'].value
        abundances_interp = pm.interpol_abundances(parameters['C/O'].value * np.ones_like(pressures), \
                                                parameters['Fe/H'].value * np.ones_like(pressures), \
                                                temperatures, \
                                                pressures,
                                                Pquench_carbon = pquench_C)
        MMW = abundances_interp['MMW']

    # Free chemistry abundances
    msum = 0.0
    if abundances_interp:
        for key, val in abundances_interp.items():
            msum += np.max(val)

    # Free chemistry species
    for species in line_species:
        if species.split("_R_")[0] in parameters.keys():
            if species.split('_')[0] in abundances_interp.keys():
                msum -= np.max(abundances_interp[species.split('_')[0]])
            abund = 10**parameters[species.split("_R_")[0]].value
            abundances_interp[species.split('_')[0]] = abund * np.ones_like(pressures)
            msum += abund

    # For free chemistry, need to fill with background gas (H2-He)
    # TODO use arbitrary background gas
    if not "Fe/H" in parameters.keys():
        # Check to make sure we're using free chemistry
        # Whatever's left is H2 and He
        if 'H2' in parameters.keys():
            abundances_interp['H2'] = 10**parameters['H2'].value * np.ones_like(pressures)
        else:
            abundances_interp['H2'] = 0.766 * (1.0-msum) * np.ones_like(pressures)
        if 'He' in  parameters.keys():
            abundances_interp['He'] = 10**parameters['He'].value * np.ones_like(pressures)
        else:
            abundances_interp['He'] = 0.234 * (1.0-msum) * np.ones_like(pressures)

        # Imposing strict limit on msum to ensure H2 dominated composition
        if msum > 1.0:
            print(f"Abundance sum > 1.0, msum={msum}")
            return None,None,None,None
        MMW = calc_MMW(abundances_interp)

    # Prior check all input params
    clouds = {}
    Pbases = {}
    abundances = {}
    for cloud in cloud_species:
        cname = cloud.split("_")[0]
        if 'use_easychem' in parameters.keys():
            # AMR CANNOT BE USED WITH EASYCHEM RIGHT NOW
            abundances[cname] = abundances_interp[cname]
            continue

        if "eq_scaling_"+cname in parameters.keys():
            # equilibrium cloud abundance
            Xcloud= fc.return_cloud_mass_fraction(cloud, parameters['Fe/H'].value, parameters['C/O'].value)
            # Scaled by a constant factor
            clouds[cname] = 10**parameters['eq_scaling_'+cname].value*Xcloud
        else:
            # Free cloud abundance
            clouds[cname] = 10**parameters['log_X_cb_'+cname].value

        # Free cloud bases
        if 'log_Pbase_'+cname in parameters.keys():
            Pbases[cname] = 10**parameters['log_Pbase_'+cname].value
        elif 'Pbase_'+cname in parameters.keys():
            Pbases[cname] = parameters['Pbase_'+cname].value
        # Equilibrium locations
        elif 'Fe/H' in parameters.keys():
            Pbases[cname] = fc.simple_cdf(cname, 
                                          pressures, 
                                          temperatures,
                                          parameters['Fe/H'].value, 
                                          parameters['C/O'].value, 
                                          np.mean(MMW))
        else:
            Pbases[cname] = fc.simple_cdf_free(cname,
                                            pressures,
                                            temperatures,
                                            10**parameters['log_X_cb_'+cname].value,
                                            MMW[0])
    # Find high resolution pressure grid and indices
    if AMR:
        press_use, small_index = fixed_length_amr(np.array(list(Pbases.values())),
                                                  pressures,
                                                  parameters['pressure_scaling'].value,
                                                  parameters['pressure_width'].value)
    else :
        small_index = np.linspace(0,pressures.shape[0]-1,pressures.shape[0], dtype = int)
    fseds = {}
    if not 'use_easychem' in parameters.keys():
        for cloud in cp.copy(cloud_species):
            cname = cloud.split('_')[0]
            # Set up fseds per-cloud
            if 'fsed_'+cname in parameters.keys():
                fseds[cname] = parameters['fsed_'+cname].value
            else:
                fseds[cname] = parameters['fsed'].value

            abundances[cname] = np.zeros_like(temperatures)
            abundances[cname][pressures < Pbases[cname]] = \
                            clouds[cname] *\
                            (pressures[pressures <= Pbases[cname]]/\
                            Pbases[cname])**fseds[cname]
            abundances[cname] = abundances[cname][small_index]


    for species in line_species:
        sname = species.split('_')[0]

        # Depending on easychem vs interpolated and different versions of pRT
        # C2H2 is named differently.
        if sname == "C2H2":
            # might be',acetylene'
            not_found = True
            for key in abundances_interp.keys():
                if sname in key:
                    sname = key
                    not_found = False
                    break
            if not_found:
                continue
        if 'FeH' in species:
            # Magic factor for FeH opacity - off by factor of 2
            abunds_change_rainout = cp.copy(abundances_interp[sname]/2.)
            if 'Fe(c)' in Pbases.keys() and not 'use_easychem' in parameters.keys():
                index_ro = pressures < Pbases['Fe(c)'] # Must have iron cloud
                abunds_change_rainout[index_ro] = 0.
            abundances[species] = abunds_change_rainout[small_index]
        abundances[species] = abundances_interp[sname][small_index]
    abundances['H2'] = abundances_interp['H2'][small_index]
    abundances['He'] = abundances_interp['He'][small_index]
    return abundances, MMW, small_index, Pbases

def get_easychem_abundances(pressures, temperatures, line_species, cloud_species, parameters, AMR = False):

    metallicity = None
    carbon_to_oxygen = None

    if "Fe/H" in parameters.keys():
        metallicity = parameters["Fe/H"].value
    if "C/O" in parameters.keys():
        carbon_to_oxygen = parameters["C/O"].value

    default_remove_condensates = ['Mg2SiO4(c)',
                                 'Mg2SiO4(L)',
                                 'MgAl2O4(c)',
                                 'FeO(c)',
                                 'Fe2SiO4(c)',
                                 'TiO2(c)',
                                 'TiO2(L)',
                                 'H3PO4(c)',
                                 'H3PO4(L)',
                                 'H2O(L)',
                                 'H2O(c)',]

    exo = ec.ExoAtmos(atoms=None,
                  reactants=None,
                  atomAbunds=None,
                  thermofpath=None,
                  feh=metallicity,
                  co=carbon_to_oxygen)

    reactants = exo.reactants.copy()

    for condensate in default_remove_condensates:
        reactants = np.delete(reactants, np.argwhere(reactants == condensate))
    exo.updateReactants(reactants)

    exo.feh = metallicity
    exo._updateFEH()

    # Set atomic abundances from parameters
    for key,val in parameters.items():
        if key not in exo.atoms:
            continue
        itemindex = np.where(exo.atoms == key)
        exo.atomAbunds[itemindex] = 10**val.value


    exo.solve(pressures, temperatures)
    abundances = exo.result_mass()
    pressures_goal = pressures.reshape(-1)
    if 'log_pquench' in parameters.keys():
        Pquench_carbon = 10**parameters['log_pquench'].value
        if Pquench_carbon > np.min(pressures_goal):
            q_index = min(np.searchsorted(pressures_goal, Pquench_carbon),
                            int(len(pressures_goal))-1)

            methane_abb = abundances['CH4']
            methane_abb[pressures_goal < Pquench_carbon] = \
                abundances['CH4'][q_index]
            abundances['CH4'] = methane_abb

            co_abb = abundances['CO']
            co_abb[pressures_goal < Pquench_carbon] = \
                abundances['CO'][q_index]
            abundances['CO'] = co_abb

            h2o_abb = abundances['H2O']
            h2o_abb[pressures_goal < Pquench_carbon] = \
                abundances['H2O'][q_index]
            abundances['H2O'] = h2o_abb
    abundances["MMW"] = exo.mmw
    return abundances

def update_atom_abundances(atom_abundances_list,atom_name_list,parameters):
    modif_abundances = atom_abundances_list.copy()
    for i_spec, name in enumerate(atom_name_list):
            if name in parameters.keys():
                modif_abundances[i_spec] = 10**parameters[name].value
    return modif_abundances

def get_exoatmos_abundances(pressures, temperatures, parameters, AMR = False):
    exo = parameters['exoatmos'].value
    metallicity = None

    if "Fe/H" in parameters.keys():
        metallicity = parameters["Fe/H"].value
        exo.Z = metallicity
    # Set atomic abundances from parameters
    atom_abundances = exo._atomAbunds.copy()
    update_abunds = update_atom_abundances(atom_abundances, exo.atoms, parameters)
    exo.updateAtomAbunds(update_abunds)
    exo.solve(pressures, temperatures)
    abundances = exo.result_mass()

    pressures_goal = pressures.reshape(-1)
    if 'log_pquench' in parameters.keys():
        Pquench_carbon = 10**parameters['log_pquench'].value
        if Pquench_carbon > np.min(pressures_goal):
            q_index = min(np.searchsorted(pressures_goal, Pquench_carbon),
                            int(len(pressures_goal))-1)

            methane_abb = abundances['CH4']
            methane_abb[pressures_goal < Pquench_carbon] = \
                abundances['CH4'][q_index]
            abundances['CH4'] = methane_abb

            co_abb = abundances['CO']
            co_abb[pressures_goal < Pquench_carbon] = \
                abundances['CO'][q_index]
            abundances['CO'] = co_abb

            h2o_abb = abundances['H2O']
            h2o_abb[pressures_goal < Pquench_carbon] = \
                abundances['H2O'][q_index]
            abundances['H2O'] = h2o_abb
    abundances["MMW"] = exo.mmw
    return abundances

def setup_exoatmos():
    exo = ec.ExoAtmos()
    # In this version, we initialize the abundances to solar
    # and fit in terms of [C/H], [O/H], and [M/H] where M is everything besides C, O, H, and He.
    exo.Z = 0.0
    exo.co = 0.55

    reactants = exo.reactants.copy()
    default_remove_condensates = ['Mg2SiO4(c)',
                                 'TiO(c)',
                                 'TiO(L)',
                                 'MgAl2O4(c)',
                                 'FeO(c)',
                                 'Fe2SiO4(c)',
                                 'TiO2(c)',
                                 'TiO2(L)',
                                 'H3PO4(c)',
                                 'H3PO4(L)']
    
    for condensate in default_remove_condensates:
        reactants = np.delete(reactants, np.argwhere(reactants == condensate))
    exo.updateReactants(reactants)

    # add FeH
    #reactants = np.append(reactants,'FeH')
    exo.updateReactants(reactants)
    return exo


"""
This module contains a set of useful functions that don't really fit anywhere
else. This includes flux conversions, prior functions, mean molecular weight
calculations, transforms from mass to number fractions, and fits file output.
"""
import os
import sys

# To not have numpy start parallelizing on its own
os.environ["OMP_NUM_THREADS"] = "1"
from scipy.special import erfcinv, gamma
import numpy as np
import math as math
from molmass import Formula
from typing import Tuple
from petitRADTRANS import nat_cst as nc
from petitRADTRANS.physics import linear_spline_profile,cubic_spline_profile
# import threading, subprocess


SQRT2 = math.sqrt(2.)


#################
# Flux scaling
#################
def surf_to_meas(flux, p_rad, dist):
    """
    surf_to_meas
    Convert from emission flux to measured flux at earth
    Args:
        flux : numpy.ndarray
            Absolute flux value or spectrum as emitted by a source of radius p_rad
        p_rad : float
            Planet radius, in same units as dist
        dist : float
            Distance to the object, in the same units as p_rad
    Returns:
        m_flux : numpy.ndarray
            Apparent flux
    """

    m_flux = flux * p_rad ** 2 / dist ** 2
    return m_flux


def freq_to_micron(frequency):
    return nc.c/frequency/1e-4


def fnu_to_flambda(wlen, spectrum):
    f_lambda = spectrum*nc.c/wlen**2.
    # convert to flux per m^2 (from flux per cm^2) cancels with step below
    #f_lambda = f_lambda * 1e4
    # convert to flux per micron (from flux per cm) cancels with step above
    #f_lambda = f_lambda * 1e-4
    # convert from ergs to Joule
    f_lambda = f_lambda * 1e-7
    return f_lambda


def spectrum_cgs_to_si(frequency, spectrum):
    wlen = freq_to_micron(frequency)
    f_lambda = fnu_to_flambda(wlen*1e-4, spectrum)
    return wlen, f_lambda


#################
# Prior Functions
#################
# Stolen from https://github.com/JohannesBuchner/MultiNest/blob/master/src/priors.f90
def log_prior(cube, lx1, lx2):
    return 10 ** (lx1 + cube * (lx2 - lx1))


def uniform_prior(cube, x1, x2):
    return x1 + cube * (x2 - x1)


def gaussian_prior(cube, mu, sigma):
    return mu + sigma * SQRT2 * erfcinv(2.0 * (1.0 - cube))
    # return -(((cube-mu)/sigma)**2.)/2.


def log_gaussian_prior(cube, mu, sigma):
    bracket = sigma * sigma + sigma * SQRT2 * erfcinv(2.0 * cube)
    return mu * np.exp(bracket)


def delta_prior(cube, x1, x2):
    return x1

def inverse_gamma_prior(cube, a, b):
    return ((b**a)/gamma(a)) * (1/cube)**(a+1) * np.exp(-b/cube)

# Sanity checks on parameter ranges
def b_range(x, b):
    if x > b:
        return -np.inf
    else:
        return 0.


def a_b_range(x, a, b):
    if x < a:
        return -np.inf
    elif x > b:
        return -np.inf
    else:
        return 0.


########################
# Mean Molecular Weights
########################

def getMM(species):
    """
    Get the molecular mass of a given species.

    This function uses the molmass package to
    calculate the mass number for the standard
    isotope of an input species. If all_iso
    is part of the input, it will return the
    mean molar mass.

    Args:
        species : string
            The chemical formula of the compound. ie C2H2 or H2O
    Returns:
        The molar mass of the compound in atomic mass units.
    """
    e_molar_mass = 5.4857990888e-4  # (g.mol-1) e- molar mass (source: NIST CODATA)

    if species == 'e-':
        return e_molar_mass
    elif species == 'H-':
        return Formula('H').mass + e_molar_mass

    name = species.split("_")[0]
    name = name.split(',')[0]
    f = Formula(name)

    if "all_iso" in species:
        return f.mass

    return f.isotope.massnumber


def calc_MMW(abundances):
    """
    calc_MMW
    Calculate the mean molecular weight in each layer.

    Args:
        abundances : dict
            dictionary of abundance arrays, each array must have the shape of the pressure array used in pRT,
            and contain the abundance at each layer in the atmosphere.
    """
    mmw = sys.float_info.min * np.ones_like(abundances[list(abundances.keys())[0]]) # prevent division by 0

    for key in abundances.keys():
        # exo_k resolution
        spec = key.split("_R_")[0]
        mmw += abundances[key] / getMM(spec)

    return 1.0 / mmw


def get_MMW_from_mfrac(m_frac):
    """
    wraps calc_MMW
    """

    return calc_MMW(m_frac)

def get_MMW_from_nfrac(n_frac):
    """
    Calculate the mean molecular weight from a number fraction

    Args:
        n_frac : dict
            A dictionary of number fractions
    """

    mass = np.zeros_like(n_frac[list(n_frac.keys())[0]])
    for key, value in n_frac.items():
        spec = key.split("_R_")[0]
        mass += value * getMM(spec)
    return mass


def mass_to_number(m_frac):
    """
    Convert mass fractions to number fractions

    Args:
        m_frac : dict
            A dictionary of mass fractions
    """

    n_frac = {}
    mmw = get_MMW_from_mfrac(m_frac)

    for key, value in m_frac.items():
        spec = key.split("_R_")[0]
        n_frac[key] = value / getMM(spec) * mmw
    return n_frac


def number_to_mass(n_fracs):
    """
    Convert number fractions to mass fractions

    Args:
        n_fracs : dict
            A dictionary of number fractions
    """

    m_frac = {}
    mmw = get_MMW_from_nfrac(n_fracs)

    for key, value in n_fracs.items():
        spec = key.split("_R_")[0]
        m_frac[key] = value * getMM(spec) / mmw
    return m_frac

def teff_calc(waves,model,dist=1.0,r_pl=1.0):
    """
    This function takes in the wavelengths and flux of a model
    in units of W/m2/micron and calculates the effective temperature
    by integrating the model and using the stefan boltzmann law.
    Args:
        waves : numpy.ndarray
            Wavelength grid in units of micron
        model : numpy.ndarray
            Flux density grid in units of W/m2/micron
        dist : Optional(float)
            Distance to the object. Must have same units as r_pl
        r_pl : Optional(float)
            Object radius. Must have same units as dist
    """
    import astropy.units as u
    import astropy.constants as c
    def integ(waves,model):
        return np.sum(model[:-1]*((dist/r_pl)**2.)*(u.W/u.m**2/u.micron)* np.diff(waves)*u.micron)

    energy = integ(waves,model)
    #print(energy)
    summed = ((energy /c.sigma_sb))
    #print(summed)
    return (summed.value)**0.25


def bin_species_exok(species, resolution):
    """
    This function uses exo-k to bin the c-k table of a
    single species to a desired (lower) spectral resolution.

    Args:
        species : string
            The name of the species
        resolution : int
            The desired spectral resolving power.
    """
    from petitRADTRANS import Radtrans
    prt_path = os.environ.get("pRT_input_data_path")
    atmosphere = Radtrans(line_species = species,
                            wlen_bords_micron = [0.1, 251.])
    ck_path = prt_path + '/opacities/lines/corr_k/'
    print("Saving to " + ck_path)
    print("Resolution: ", resolution)

    masses = {}

    for spec in species:
        masses[spec.split('_')[0]] = getMM(spec)

    atmosphere.write_out_rebin(
        resolution=int(resolution),
        path=ck_path,
        species=species,
        masses=masses
    )

def compute_gravity(parameters):
    gravity = -np.inf
    R_pl = -np.inf
    if 'log_g' in parameters.keys() and 'mass' in parameters.keys():
        gravity = 10**parameters['log_g'].value
        R_pl = np.sqrt(nc.G*parameters['mass'].value/gravity)
    elif 'log_g' in parameters.keys():
        gravity= 10**parameters['log_g'].value
        R_pl = parameters['R_pl'].value
    elif 'mass' in parameters.keys():
        R_pl = parameters['R_pl'].value
        gravity = nc.G * parameters['mass'].value/R_pl**2
    else:
        print("Pick two of log_g, R_pl and mass priors!")
        sys.exit(5)
    return gravity, R_pl

def set_resolution(lines,abundances,resolution):
    """
    deprecated
    """
    # Set correct key names in abundances for pRT, with set resolution
    # Only needed for free chemistry retrieval
    #print(lines)
    #print(abundances)
    if resolution is None:
        return abundances
    for line in lines:
        abundances[line] = abundances[line.split("_R_"+str(resolution))[0]]
        del abundances[line.split("_R_"+str(resolution))]
    return abundances


def fixed_length_amr(p_clouds, pressures, scaling = 10, width = 3):
    r"""This function takes in the cloud base pressures for each cloud,
    and returns an array of pressures with a high resolution mesh
    in the region where the clouds are located.

    Author:  Francois Rozet.

    The output length is always
        len(pressures[::scaling]) + len(p_clouds) * width * (scaling - 1)

    Args:
        P_clouds : numpy.ndarray
            The cloud base pressures in bar
        press : np.ndarray
            The high resolution pressure array.
        scaling : int
            The factor by which the low resolution pressure array is scaled
        width : int
            The number of low resolution bins to be replaced for each cloud layer.
    """

    length = len(pressures)
    cloud_indices = np.searchsorted(pressures, np.asarray(p_clouds))

    # High resolution intervals
    def bounds(center: int, width: int) -> Tuple[int, int]:
        upper = min(center + width // 2, length)
        lower = max(upper - width, 0)
        return lower, lower + width

    intervals = [bounds(idx, scaling * width) for idx in cloud_indices]

    # Merge intervals
    while True:
        intervals, stack = sorted(intervals), []

        for interval in intervals:
            if stack and stack[-1][1] >= interval[0]:
                last = stack.pop()
                interval = bounds(
                    (last[0] + max(last[1], interval[1]) + 1) // 2,
                    last[1] - last[0] + interval[1] - interval[0],
                )

            stack.append(interval)

        if len(intervals) == len(stack):
            break
        intervals = stack

    # Intervals to indices
    indices = [np.arange(0, length, scaling)]

    for interval in intervals:
        indices.append(np.arange(*interval))

    indices = np.unique(np.concatenate(indices))

    return pressures[indices], indices

########################
# File Formatting
########################
def fits_output(wavelength, spectrum, covariance, object, output_dir="", correlation=None):  # TODO arg object needs to be renamed
    """
    Generate a fits file that can be used as an input to a pRT retrieval.

    Args:
        wavelength : numpy.ndarray
            The wavelength bin centers in micron. dim(N)
        spectrum : numpy.ndarray
            The flux density in W/m2/micron at each wavelength bin. dim(N)
        covariance : numpy.ndarray
            The covariance of the flux in (W/m2/micron)^2 dim(N,N)
        object : string
            The name of the object, used for file naming.
        output_dir : string
            The parent directory of the output file.
        correlation : numpy.ndarray
            The correlation matrix of the flux points (See Brogi & Line 2018, https://arxiv.org/pdf/1811.01681.pdf)

    Returns:
        hdul : astropy.fits.HDUlist
            The HDUlist object storing the spectrum.
    """

    from astropy.io import fits
    primary_hdu = fits.PrimaryHDU([])
    primary_hdu.header['OBJECT'] = object
    c1 = fits.Column(name="WAVELENGTH", array=wavelength, format='D', unit="micron")
    c2 = fits.Column(name="FLUX", array=spectrum, format='D', unit="W/m2/micron")
    c3 = fits.Column(name="COVARIANCE", array=covariance, format=str(covariance.shape[0]) + 'D', unit="[W/m2/micron]^2")
    if correlation is not None:
        c4 = fits.Column(name="CORRELATION", array=correlation, format=str(correlation.shape[0]) + 'D', unit=" - ")
    columns = [c1, c2, c3, c4]
    table_hdu = fits.BinTableHDU.from_columns(columns, name='SPECTRUM')
    hdul = fits.HDUList([primary_hdu, table_hdu])
    outstring = os.path.join(output_dir, object + "_spectrum.fits")
    hdul.writeto(outstring, overwrite=True, checksum=True, output_verify='exception')
    return hdul
